/*!
 * @project        The CERN Tape Archive (CTA)
 * @brief          CTA Frontend Failed Request Ls stream implementation
 * @copyright      Copyright 2019 CERN
 * @license        This program is free software: you can redistribute it and/or modify
 *                 it under the terms of the GNU General Public License as published by
 *                 the Free Software Foundation, either version 3 of the License, or
 *                 (at your option) any later version.
 *
 *                 This program is distributed in the hope that it will be useful,
 *                 but WITHOUT ANY WARRANTY; without even the implied warranty of
 *                 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *                 GNU General Public License for more details.
 *
 *                 You should have received a copy of the GNU General Public License
 *                 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <xroot_plugins/XrdCtaStream.hpp>
#include <xroot_plugins/XrdSsiCtaRequestMessage.hpp>


namespace cta { namespace xrd {

/*!
 * Stream object which implements "failedrequest ls" command.
 */
class FailedRequestLsStream : public XrdCtaStream
{
public:
  /*!
   * Constructor
   *
   * @param[in]    requestMsg    RequestMessage containing command-line arguments
   * @param[in]    catalogue     CTA Catalogue
   * @param[in]    scheduler     CTA Scheduler
   * @param[in]    schedDb       CTA ObjectStore
   * @param[in]    lc            CTA Log Context
   */
  FailedRequestLsStream(const RequestMessage &requestMsg, cta::catalogue::Catalogue &catalogue,
    cta::Scheduler &scheduler, OStoreDB &schedDb, log::LogContext &lc);

private:
  /*!
   * Can we close the stream?
   */
  virtual bool isDone() const {
    return m_isSummary ? m_isSummaryDone
                       : !isArchiveJobs() && !isRetrieveJobs();
  }

  bool isArchiveJobs() const {
    return m_archiveQueueItorPtr && !m_archiveQueueItorPtr->end();
  }

  bool isRetrieveJobs() const {
    return m_retrieveQueueItorPtr && !m_retrieveQueueItorPtr->end();
  }

  /*!
   * Fill the buffer
   */
  virtual int fillBuffer(XrdSsiPb::OStreamBuffer<Data> *streambuf);

  /*!
   * Add a record to the stream
   *
   * @param[in]    streambuf      XRootD SSI stream object to push records to
   * @param[in]    item           Archive or Retrieve Job (used for template specialisation)
   *
   * @retval    true     Stream buffer is full and ready to send
   * @retval    false    Stream buffer is not full
   */
  template<typename QueueType>
  bool pushRecord(XrdSsiPb::OStreamBuffer<Data> *streambuf, const QueueType &item);

  /*!
   * Populate the failed queue summary
   *
   * @param[in]    streambuf         XRootD SSI stream object to push records to
   */
  void GetBuffSummary(XrdSsiPb::OStreamBuffer<Data> *streambuf);


  std::unique_ptr<OStoreDB::ArchiveQueueItor_t>  m_archiveQueueItorPtr;     //!< Archive Queue Iterator
  std::unique_ptr<OStoreDB::RetrieveQueueItor_t> m_retrieveQueueItorPtr;    //!< Retrieve Queue Iterator
  bool m_isSummary;                                                         //!< Show only summary of items in the failed queues
  bool m_isSummaryDone;                                                     //!< Summary has been sent
  bool m_isLogEntries;                                                      //!< Show failure log messages (verbose)
  log::LogContext &m_lc;                                                    //!< Reference to CTA Log Context

  static constexpr const char* const LOG_SUFFIX  = "FailedRequestLsStream"; //!< Identifier for SSI log messages
};


FailedRequestLsStream::FailedRequestLsStream(const RequestMessage &requestMsg,
  cta::catalogue::Catalogue &catalogue, cta::Scheduler &scheduler, OStoreDB &schedDb,
  log::LogContext &lc) :
    XrdCtaStream(catalogue, scheduler),
    m_isSummary(requestMsg.has_flag(admin::OptionBoolean::SUMMARY)),
    m_isSummaryDone(false),
    m_isLogEntries(requestMsg.has_flag(admin::OptionBoolean::SHOW_LOG_ENTRIES)),
    m_lc(lc)
{
  using namespace cta::admin;

  XrdSsiPb::Log::Msg(XrdSsiPb::Log::DEBUG, LOG_SUFFIX, "FailedRequestLsStream() constructor");

  if(m_isLogEntries && m_isSummary) {
    throw cta::exception::UserError("--log and --summary are mutually exclusive");
  }

  auto tapepool     = requestMsg.getOptional(OptionString::TAPE_POOL);
  auto vid          = requestMsg.getOptional(OptionString::VID);
  bool justarchive  = requestMsg.has_flag(OptionBoolean::JUSTARCHIVE)  || tapepool;
  bool justretrieve = requestMsg.has_flag(OptionBoolean::JUSTRETRIEVE) || vid;

  if(justarchive && justretrieve) {
    throw cta::exception::UserError("--justarchive/--tapepool and --justretrieve/--vid options are mutually exclusive");
  }

  if(!justretrieve)
    m_archiveQueueItorPtr.reset(schedDb.getArchiveJobItorPtr(tapepool ? *tapepool : "", objectstore::JobQueueType::FailedJobs));
  if(!justarchive)
    m_retrieveQueueItorPtr.reset(schedDb.getRetrieveJobItorPtr(vid ? *vid : "", objectstore::JobQueueType::FailedJobs));
}


/*!
 * pushRecord ArchiveJob specialisation
 */
template<> bool FailedRequestLsStream::
pushRecord(XrdSsiPb::OStreamBuffer<Data> *streambuf, const common::dataStructures::ArchiveJob &item)
{
  Data record;

  record.mutable_frls_item()->set_request_type(admin::RequestType::ARCHIVE_REQUEST);
  record.mutable_frls_item()->set_tapepool(m_archiveQueueItorPtr->qid());
  record.mutable_frls_item()->set_copy_nb(item.copyNumber);
  record.mutable_frls_item()->mutable_requester()->set_username(item.request.requester.name);
  record.mutable_frls_item()->mutable_requester()->set_groupname(item.request.requester.group);
  record.mutable_frls_item()->mutable_af()->mutable_df()->set_path(item.request.diskFileInfo.path);

  if(m_isLogEntries) {
    *record.mutable_frls_item()->mutable_failurelogs() = { item.failurelogs.begin(), item.failurelogs.end() };
  }
  return streambuf->Push(record);
}


/*!
 * pushRecord RetrieveJob specialisation
 */
template<> bool FailedRequestLsStream::
pushRecord(XrdSsiPb::OStreamBuffer<Data> *streambuf, const common::dataStructures::RetrieveJob &item)
{
  auto &vid = m_retrieveQueueItorPtr->qid();

  Data record;

  record.mutable_frls_item()->set_request_type(admin::RequestType::RETRIEVE_REQUEST);
  record.mutable_frls_item()->mutable_tf()->set_vid(vid);
  record.mutable_frls_item()->set_copy_nb(item.tapeCopies.at(vid).first);
  record.mutable_frls_item()->mutable_requester()->set_username(item.request.requester.name);
  record.mutable_frls_item()->mutable_requester()->set_groupname(item.request.requester.group);
  record.mutable_frls_item()->mutable_af()->mutable_df()->set_path(item.request.diskFileInfo.path);

  if(m_isLogEntries) {
    *record.mutable_frls_item()->mutable_failurelogs() = { item.failurelogs.begin(), item.failurelogs.end() };
  }
  return streambuf->Push(record);
}


int FailedRequestLsStream::fillBuffer(XrdSsiPb::OStreamBuffer<Data> *streambuf) {
  if(m_isSummary) {
    // Special handling for -S (Summary) option
    GetBuffSummary(streambuf);
  } else {
    // List failed archive requests
    if(isArchiveJobs()) {
      for(bool is_buffer_full = false; !m_archiveQueueItorPtr->end() && !is_buffer_full; ++*m_archiveQueueItorPtr) {
        is_buffer_full = pushRecord(streambuf, **m_archiveQueueItorPtr);
      }
    }
    // List failed retrieve requests
    if(isRetrieveJobs()) {
      for(bool is_buffer_full = false; !m_retrieveQueueItorPtr->end() && !is_buffer_full; ++*m_retrieveQueueItorPtr) {
        is_buffer_full = pushRecord(streambuf, **m_retrieveQueueItorPtr);
      }
    }
  }
  return streambuf->Size();
}



void FailedRequestLsStream::
GetBuffSummary(XrdSsiPb::OStreamBuffer<Data> *streambuf)
{
  SchedulerDatabase::JobsFailedSummary archive_summary;
  SchedulerDatabase::JobsFailedSummary retrieve_summary;

  if(isArchiveJobs()) {
    Data record;
    archive_summary = m_scheduler.getArchiveJobsFailedSummary(m_lc);
    record.mutable_frls_summary()->set_request_type(admin::RequestType::ARCHIVE_REQUEST);
    record.mutable_frls_summary()->set_total_files(archive_summary.totalFiles);
    record.mutable_frls_summary()->set_total_size(archive_summary.totalBytes);
    streambuf->Push(record);
  }
  if(isRetrieveJobs()) {
    Data record;
    retrieve_summary = m_scheduler.getRetrieveJobsFailedSummary(m_lc);
    record.mutable_frls_summary()->set_request_type(admin::RequestType::RETRIEVE_REQUEST);
    record.mutable_frls_summary()->set_total_files(retrieve_summary.totalFiles);
    record.mutable_frls_summary()->set_total_size(retrieve_summary.totalBytes);
    streambuf->Push(record);
  }
  if(isArchiveJobs() && isRetrieveJobs()) {
    Data record;
    record.mutable_frls_summary()->set_request_type(admin::RequestType::TOTAL);
    record.mutable_frls_summary()->set_total_files(archive_summary.totalFiles + retrieve_summary.totalFiles);
    record.mutable_frls_summary()->set_total_size(archive_summary.totalBytes + retrieve_summary.totalBytes);
    streambuf->Push(record);
  }

  m_isSummaryDone = true;
}

}} // namespace cta::xrd
